/* Copyright (C) 2009 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <alloca.h>

#include "huffman.h"

/* Construct a huffman tree representing the codes starting with the
   given prefix. Returns the number of used nodes. */
/* Highly unoptimized. */
static unsigned
huffman_fill_table(struct huffman *table,
		   unsigned n,
		   const unsigned char *length,
		   const unsigned short *code,
		   unsigned prefix_length, unsigned prefix)
{
  unsigned k;
  for (k = 0; k < n; k++)
    {
      if (length[k] && length[k] >= prefix_length
	  && (code[k] >> (length[k] - prefix_length)) == prefix)
	{
	  if (length[k] == prefix_length)
	    {
	      /* A leaf node */
	      table->left = -1;
	      table->right = k;
	      return 1;
	    }
	  else
	    {
	      unsigned left
		= huffman_fill_table(table + 1,
				     n, length, code,
				     prefix_length+1,
				     prefix << 1);
	      unsigned right
		= huffman_fill_table(table + 1 + left,
				     n, length, code,
				     prefix_length + 1,
				     (prefix << 1) + 1);
						  
	      /* Always 1 (or zero)? Could be compressed. */
	      table->left = left > 0 ? 1 : 0;
	      table->right = right > 0 ? 1 + left : 0;
	      return 1 + left + right;
	    }
	}
    }

  /* No codes in interval. Represent as a zero pointer. */
  return 0;
}

/* FIXME: What's the right value? */
#define MAX_LENGTH 15

int
huffman_table(struct huffman *table,
	      unsigned n, const unsigned char *length)
{
  unsigned char count[MAX_LENGTH+1];
  unsigned next[MAX_LENGTH+1];
  unsigned k;
  unsigned bits;
  unsigned short *code;
  unsigned size;

  memset(count, 0, sizeof(count));

  for (k = 0; k < n; k++)
    {
      if (length[k] > MAX_LENGTH)
	return 0;
      count[length[k]]++;
    }

  /* RFC 1951 seem to have an off-by-one error. */
  next[1] = 0;
  for (bits = 2; bits <= MAX_LENGTH; bits++)
    {
      next[bits] = (next[bits-1] + count[bits-1]) << 1;

      /* Make sure code fits in given size. */
      if (next[bits] + count[bits] > (1 << bits))
	return 0;
    }
  
  code = alloca(n * sizeof(*code));

  for (k = 0; k < n; k++)
    {
      if (length[k] > 0)
	code[k] = next[length[k]]++;

      /* FIXME: What about the reverse mapping? */
    }

  size = huffman_fill_table(table, n, length, code, 0, 0);
  assert(size < HUFFMAN_SIZE(n));

  /* Sanity check */
  for (k = 0; k < size; k++)
    {
      assert(table[k].left == -1
	     || (k + table[k].left < size
		 && k + table[k].right < size));
    }

  return 1;
}

const struct huffman *
huffman_decode_bit(const struct huffman *table, unsigned bit, int *value)
{
  unsigned incr = bit ? table->right : table->left;

  if (!incr)
    {
      *value = -1;
      return NULL;
    }

  table += incr;

  if (table->left == -1)
    {
      *value = table->right;
      return NULL;
    }
  else
    return table;
}
